import os
import random
import tarfile
from copy import deepcopy
from typing import Callable, Optional

import numpy as np
import requests
import torch
import torchvision.transforms as transforms
from PIL import Image
from tqdm import tqdm

from spuco.datasets import (TEST_SPLIT, TRAIN_SPLIT, VAL_SPLIT,
                            BaseSpuCoDataset, SourceData)
from spuco.utils.random_seed import seed_randomness

# Constants
DOWNLOAD_URL = "https://ucla.box.com/shared/static/u3dkllvq7a2py4x443pqul295wwd549e"
DATASET_NAME = "spuco_dogs"
SMALL_DOGS = "small_dogs"
BIG_DOGS = "big_dogs"
INDOOR = "indoor"
OUTDOOR = "outdoor"
MAJORITY_SIZE = {
    TRAIN_SPLIT: 10000,
    VAL_SPLIT: 500,
    TEST_SPLIT: 500,
}
MINORITY_SIZE = {
    TRAIN_SPLIT: 500,
    VAL_SPLIT: 25,
    TEST_SPLIT: 500,
}

class SpuCoDogs(BaseSpuCoDataset):
    """
    Subset of SpuCoAnimals only including Dog classes.
    """

    def __init__(
        self,
        root: str,
        download: bool = True,
        label_noise: float = 0.0,
        split: str = TRAIN_SPLIT,
        transform: Optional[Callable] = None,
        verbose: bool = False
    ):
        """
        Initializes the dataset.

        :param root: Root directory of the dataset.
        :type root: str

        :param download: Whether to download the dataset.
        :type download: bool, optional

        :param label_noise: The amount of label noise to apply.
        :type label_noise: float, optional

        :param split: The split of the dataset.
        :type split: str, optional

        :param transform: Optional transform to be applied to the data.
        :type transform: Callable, optional

        :param verbose: Whether to print verbose information during dataset initialization.
        :type verbose: bool, optional
        """

        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        super().__init__(
            root=root, 
            split=split,
            transform=transform,
            num_classes=2,
            verbose=verbose
        )
        self.download = download
        self.label_noise = label_noise
        self.base_transform = transforms.Compose([
            transforms.Resize((256, 256)),
            transforms.CenterCrop((224, 224)),
            transforms.ToTensor()
        ])
        
    def load_data(self) -> SourceData:
        """
        Loads SpuCoDogs and sets spurious labels, label noise.

        :return: The spurious correlation dataset.
        :rtype: SourceData, List[int], List[int]
        """
        
        self.dset_dir = os.path.join(self.root, DATASET_NAME, self.split)
        if not os.path.exists(self.dset_dir):
            if not self.download:
                raise RuntimeError(f"Dataset not found {self.dset_dir}, run again with download=True")
            self._download_data()
            self._untar_data()
            os.remove(self.filename)
            
        try:
            self.data = SourceData()
            
            # Small Dogs - Indoor
            small_dogs_indoor = os.listdir(os.path.join(self.dset_dir, f"{SMALL_DOGS}/{INDOOR}"))
            self.data.X.extend([str(os.path.join(self.dset_dir, f"{SMALL_DOGS}/{INDOOR}", x)) for x in small_dogs_indoor])
            self.data.labels.extend([0] * len(small_dogs_indoor))
            self.data.spurious.extend([0] * len(small_dogs_indoor))
            assert len(small_dogs_indoor) == MAJORITY_SIZE[self.split], f"Dataset corrupted or missing files. Expected {MAJORITY_SIZE[self.split]} files got {len(small_dogs_indoor)}"
            
            # Small Dogs - Outdoor
            small_dogs_outdoor = os.listdir(os.path.join(self.dset_dir, f"{SMALL_DOGS}/{OUTDOOR}"))
            self.data.X.extend([str(os.path.join(self.dset_dir, f"{SMALL_DOGS}/{OUTDOOR}", x)) for x in small_dogs_outdoor])
            self.data.labels.extend([0] * len(small_dogs_outdoor))
            self.data.spurious.extend([1] * len(small_dogs_outdoor))   
            assert len(small_dogs_outdoor) == MINORITY_SIZE[self.split], f"Dataset corrupted or missing files. Expected {MINORITY_SIZE[self.split]} files got {len(small_dogs_outdoor)}"
            
            # Big Dogs - Indoor
            big_dogs_indoor = os.listdir(os.path.join(self.dset_dir, f"{BIG_DOGS}/{INDOOR}"))
            self.data.X.extend([str(os.path.join(self.dset_dir, f"{BIG_DOGS}/{INDOOR}", x)) for x in big_dogs_indoor])
            self.data.labels.extend([1] * len(big_dogs_indoor))
            self.data.spurious.extend([0] * len(big_dogs_indoor))
            assert len(big_dogs_indoor) == MINORITY_SIZE[self.split], f"Dataset corrupted or missing files. Expected {MINORITY_SIZE[self.split]} files got {len(big_dogs_indoor)}"
            
            # Big Dogs - Outdoor
            big_dogs_outdoor = os.listdir(os.path.join(self.dset_dir, f"{BIG_DOGS}/{OUTDOOR}"))
            self.data.X.extend([str(os.path.join(self.dset_dir, f"{BIG_DOGS}/{OUTDOOR}", x)) for x in big_dogs_outdoor])
            self.data.labels.extend([1] * len(big_dogs_outdoor))
            self.data.spurious.extend([1] * len(big_dogs_outdoor)) 
            assert len(big_dogs_outdoor) == MAJORITY_SIZE[self.split], f"Dataset corrupted or missing files. Expected {MAJORITY_SIZE[self.split]} files got {len(big_dogs_outdoor)}"
            
            if self.label_noise > 0.0:
                self.data.clean_labels = deepcopy(self.data.labels)
                self.is_noisy_label = torch.zeros(len(self.data.X))
                self.is_noisy_label[torch.randperm(len(self.data.X))[:int(self.label_noise * len(self.data.X))]] = 1
                self.data.labels = [1 - label if self.is_noisy_label[i] else label for i, label in enumerate(self.data.clean_labels)]
        except:
            raise RuntimeError(f"Dataset corrupted, please delete {self.dset_dir} and run again with download=True")
            
        return self.data, list(range(2)), list(range(2))

    def _download_data(self):
        self.filename = f"{self.root}/{DATASET_NAME}.tar.gz"

        response = requests.get(DOWNLOAD_URL, stream=True)
        response.raise_for_status()

        with open(self.filename, "wb") as file:
            for chunk in tqdm(response.iter_content(chunk_size=1024), total=2593935, desc="Downloading SpuCoDogs", unit="KB"):
                file.write(chunk)
    
    def _untar_data(self):
        # Open the tar.gz file
        with tarfile.open(self.filename, "r:gz") as tar:
            # Extract all files to the specified output directory
            tar.extractall(self.root)
            
    def __getitem__(self, index):
        """
        Gets an item from the dataset.

        :param index: Index of the item to get.
        :type index: int
        :return: A tuple of (sample, target) where target is class_index of the target class.
        :rtype: tuple
        """
        
        image = self.base_transform(Image.open(self.load_image(self.data.X[index])))
        label = self.data.labels[index]
        if self.transform is None:
            return image, label
        else:
            return self.transform(image), label


import random
from typing import Dict, List, Optional

import numpy as np
import torch
from sklearn.metrics import silhouette_score
from tqdm import tqdm

from spuco.group_inference.cluster import ClusterAlg
from spuco.group_inference import Cluster
from spuco.utils.random_seed import seed_randomness, get_seed


class SpareInference(Cluster):
    """
    SPARE Inference: https://arxiv.org/abs/2305.18761
    """
    def __init__(
        self,
        Z: torch.Tensor,
        class_labels: Optional[List[int]] = None,
        cluster_alg: ClusterAlg = ClusterAlg.KMEANS,
        num_clusters: int = -1,
        max_clusters: int = -1,
        silhoutte_threshold: float = 0.9,
        high_sampling_power: int = 2,
        device: torch.device = torch.device("cpu"), 
        verbose: bool = False
    ):
        """
        Initializes Spare Inference.

        :param Z: The output of the network.
        :type Z: torch.Tensor
        :param class_labels: Optional list of class labels for class-wise clustering. Defaults to None.
        :type class_labels: Optional[List[int]], optional
        :param cluster_alg: The clustering algorithm to use. Defaults to ClusterAlg.KMEANS.
        :type cluster_alg: ClusterAlg, optional
        :param num_clusters: The number of clusters to create. Defaults to -1.
        :type num_clusters: int, optional
        :param max_clusters: The maximum number of clusters to consider. Defaults to -1.
        :type max_clusters: int, optional
        :param silhoutte_threshold: The silhouette threshold for determining the sampling powers. Defaults to 0.9.
        :type silhoutte_threshold: float, optional
        :param high_sampling_power: The sampling power for the low-silhouette clusters. Defaults to 2.
        :type high_sampling_power: int, optional
        :param device: The device to run the clustering on. Defaults to torch.device("cpu").
        :type device: torch.device, optional
        :param verbose: Whether to display progress and logging information. Defaults to False.
        :type verbose: bool, optional
        """
        
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)
        super().__init__(Z=Z, class_labels=class_labels, cluster_alg=cluster_alg, num_clusters=num_clusters, max_clusters=max_clusters, random_seed=get_seed(), device=device, verbose=verbose)

        self.silhouette_threshold = silhoutte_threshold
        self.high_sampling_power = high_sampling_power
    
    def infer_groups(self) -> Dict[int, List[int]]:
        """
        Infers the group partition based on the clustering results.

        :return: The group partition.
        :rtype: Dict[int, List[int]]

        :return: The sampling powers for each group.
        :rtype: List[int]
        """ 
        # Get class-wise group partitions
        cluster_partitions = [] 
        sampling_powers = []
        for class_label in tqdm(self.class_partition.keys(), disable=not self.verbose, desc="Clustering class-wise"):
            Z = self.Z[self.class_partition[class_label]]
            if self.num_clusters == -1:
                partition, silhouette = self.silhouette(Z)
            else:
                cluster_labels, partition = self.kmeans(Z, num_clusters=self.num_clusters)
                silhouette = silhouette_score(Z, cluster_labels)
            cluster_partitions.append(partition)
            if silhouette < self.silhouette_threshold:
                sampling_powers.append(self.high_sampling_power)
            else:
                sampling_powers.append(1)

        # Merge class-wise group partitions into one dictionary
        group_partition = {}
        for class_index, partition in zip(self.class_partition.keys(), cluster_partitions):
            group_partition.update(self.process_cluster_partition(partition, class_index))
        return group_partition, sampling_powers


import random as random

import numpy as np
import torch
from transformers import BertModel

from spuco.utils.random_seed import seed_randomness

# Acknowledgement to
# https://github.com/p-lambda/wilds/blob/main/examples/models/bert/bert.py


class Bert(BertModel):
    def __init__(self, config):
        """
        Initializes a BERT model.

        :param config: Configuration for the BERT model.
        :type config: transformers.configuration_bert.BertConfig
        """
        seed_randomness(random_module=random, torch_module=torch, numpy_module=np)
        super().__init__(config)
        self.d_out = config.hidden_size

    def __call__(self, x):
        """
        Forward pass of the BERT model.

        :param x: Input tensor.
        :type x: torch.Tensor

        :return: Output tensor.
        :rtype: torch.Tensor
        """
        input_ids = x[:, :, 0]
        attention_mask = x[:, :, 1]
        token_type_ids = x[:, :, 2]
        outputs = super().__call__(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids
        )[1] # get pooled output
        return outputs